
import org.sosy_lab.common.ShutdownManager;
import org.sosy_lab.common.configuration.Configuration;
import org.sosy_lab.common.configuration.InvalidConfigurationException;
import org.sosy_lab.common.log.BasicLogManager;
import org.sosy_lab.common.log.LogManager;
import org.sosy_lab.java_smt.SolverContextFactory;
import org.sosy_lab.java_smt.SolverContextFactory.Solvers;
import org.sosy_lab.java_smt.api.*;
import org.sosy_lab.java_smt.api.NumeralFormula.IntegerFormula;
import org.sosy_lab.java_smt.api.SolverContext.ProverOptions;
import java.util.List;
import java.util.ArrayList;
import java.util.Set;
import java.util.HashSet;

/**
 * A simple interpolation example.
 */
public class Interpolation {
  public static void main(String[] args)
      throws InvalidConfigurationException,
             SolverException,
             InterruptedException {
    Configuration config = Configuration.fromCmdLineArguments(args);
    LogManager logger = BasicLogManager.create(config);
    ShutdownManager shutdown = ShutdownManager.create();

    System.out.println("Solving ...");

    SolverContext context = SolverContextFactory.createSolverContext(
        config, logger, shutdown.getNotifier(), Solvers.PRINCESS);

    FormulaManager fmgr = context.getFormulaManager();

    BooleanFormulaManager bmgr = fmgr.getBooleanFormulaManager();
    IntegerFormulaManager imgr = fmgr.getIntegerFormulaManager();

    IntegerFormula x = imgr.makeVariable("x"),
                   a = imgr.makeVariable("a"),
                   b = imgr.makeVariable("b"),
                   c = imgr.makeVariable("c");

    try (InterpolatingProverEnvironment<Object> prover =
           (InterpolatingProverEnvironment<Object>)
             context.newProverEnvironmentWithInterpolation()) {

        /**

           \part[cond]          (a-2*x = 0 & -a <= 0) &
           \part[stmt1]         (2*b - a <=0 & -2*b + a -1 <=0) &
           \part[stmt2]         c-3*b-1=0
                                ->
           \part[assert]        c > a

         */

        final BooleanFormula cond =
            bmgr.and(imgr.equal(
                       imgr.subtract(a, imgr.multiply(imgr.makeNumber(2), x)),
                       imgr.makeNumber(0)),
                     imgr.lessOrEquals(
                       imgr.negate(a),
                       imgr.makeNumber(0)));
        final BooleanFormula stmt1 =
            bmgr.and(imgr.lessOrEquals(
                       imgr.subtract(imgr.multiply(imgr.makeNumber(2), b), a),
                       imgr.makeNumber(0)),
                     imgr.lessOrEquals(
                       imgr.subtract(
                         imgr.add(imgr.multiply(imgr.makeNumber(-2), b), a),
                         imgr.makeNumber(1)),
                       imgr.makeNumber(0)));
        final BooleanFormula stmt2 =
                     imgr.equal(
                       imgr.subtract(
                         imgr.subtract(c, imgr.multiply(imgr.makeNumber(3), b)),
                         imgr.makeNumber(1)),
                       imgr.makeNumber(0));
        final BooleanFormula asrt =
            bmgr.not(imgr.greaterThan(c, a));

        final Object condTag = prover.push(cond);
        final Object stmt1Tag = prover.push(stmt1);
        final Object stmt2Tag = prover.push(stmt2);
        final Object asrtTag = prover.push(asrt);
        
      if (prover.isUnsat()) {
          System.out.println("Valid implication");
          System.out.println("Interpolants:");

          List<Set<Object>> partitions = new ArrayList<>();

          partitions.add(toSet(condTag));
          partitions.add(toSet(stmt1Tag));
          partitions.add(toSet(stmt2Tag));
          partitions.add(toSet(asrtTag));
          
          List<BooleanFormula> itps = prover.getSeqInterpolants(partitions);

          for (BooleanFormula f : itps)
              System.out.println(f);
      }
    }
  }

  private static <T> Set<T> toSet(T x) {
      final Set<T> res = new HashSet<>();
      res.add(x);
      return res;
  }
}
